//////////////////////////////////////////////////////////////////////////
//	"tform.cpp"
//------------------------------------------------------------------------
//	$Id: tform.cpp 118 2004-01-14 21:27:24Z taro $
//	2004.01.14	for devtest & dvd4ram
//////////////////////////////////////////////////////////////////////////

#include	"tform.h"

#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>

#if		defined(WIN32)
#	define	I64(n)	n##i64
#	define	U64(n)	n##ui64
#elif	defined(LINUX)
#	ifndef	TEXT
#	 define	TEXT(s)	s
#	endif
#	define	I64(n)	n##ll
#	define	U64(n)	n##ull
#	define	_tcslen	strlen
#	define	_tcscpy	strcpy
#	ifndef	FALSE
#	 define	FALSE	0
#	endif
#	ifndef	TRUE
#	 define	TRUE	1
#	endif
#endif

const TCHAR	TForm::ntoa_ltbl[16] =
{
	 TEXT('0'),TEXT('1'),TEXT('2'),TEXT('3')
	,TEXT('4'),TEXT('5'),TEXT('6'),TEXT('7')
	,TEXT('8'),TEXT('9'),TEXT('a'),TEXT('b')
	,TEXT('c'),TEXT('d'),TEXT('e'),TEXT('f')
};
const TCHAR	TForm::ntoa_utbl[16] =
{
	 TEXT('0'),TEXT('1'),TEXT('2'),TEXT('3')
	,TEXT('4'),TEXT('5'),TEXT('6'),TEXT('7')
	,TEXT('8'),TEXT('9'),TEXT('A'),TEXT('B')
	,TEXT('C'),TEXT('D'),TEXT('E'),TEXT('F')
};

int		TForm::Printf(const TCHAR * form, ...)
{
	va_list		arg;
	TCHAR		buffer[BUFFER_SIZE];
	int			len;

	va_start(arg, form);
	len = TForm::format(sizeof(buffer), buffer, form, arg);
	va_end(arg);
	fputs(buffer, stdout);
	return len;
}

int		TForm::SPrintf(TCHAR * buffer, const TCHAR * form, ...)
{
	va_list		arg;
	int			len;

	va_start(arg, form);
	len = TForm::format(BUFFER_SIZE, buffer, form, arg);
	va_end(arg);
	return len;
}

int		TForm::format(unsigned size, TCHAR * buffer, const TCHAR * form, va_list arg)
{
	const TCHAR		*src = form;
	TCHAR			*dst = buffer;
	unsigned		outsize = 0;

	while ( *src && outsize < size )
	{
		if ( *src == TEXT('%') )
		{
			++src;
			if ( *src == TEXT('%') )	// "%%"
			{
				*dst++ = *src++;
				outsize += sizeof(TCHAR);
			} else
			{
				Format	fmt;
				BOOL	loop;

				fmt.left_margin  = FALSE;
				fmt.print_sign   = FALSE;
				fmt.zero_pad     = FALSE;
				fmt.sign_space   = FALSE;
				fmt.num_symbol   = FALSE;
				fmt.width        = -1;
				fmt.precision    = -1;
				fmt.part_ch      = TEXT(',');
				fmt.part_spc     = -1;
				fmt.opt          = OPT_OFF;
				fmt.type         = 0;

				// %[flags] [width] [precision] [{h | l | I64 | L}]type
				// flags
				loop = TRUE;
				while ( *src && loop )
				{
					switch ( *src )
					{
						case TEXT('-'): *src++; fmt.left_margin = TRUE; break;
						case TEXT('+'): *src++; fmt.print_sign  = TRUE; break;
						case TEXT('0'): *src++; fmt.zero_pad    = TRUE; break;
						case TEXT(' '): *src++; fmt.sign_space  = TRUE; break;
						case TEXT('#'): *src++; fmt.num_symbol  = TRUE; break;
						default:
							loop = FALSE;
							break;
					}
				}
				// width
				if ( *src == TEXT('*') )
				{
					++src;
					fmt.width = va_arg(arg, int);
				} else
				{
					unsigned	n;

					n = 0;
					while ( *src >= TEXT('0') && *src <= TEXT('9') )
						n = n * 10 + ((*src++) - TEXT('0'));
					if ( n > 0 )
						fmt.width = n;
				}
				// precision
				if ( *src == TEXT('.') )
				{
					++src;
					if ( *src == TEXT('*') )
					{
						++src;
						fmt.precision = va_arg(arg, int);
					} else
					{
						unsigned	n;

						n = 0;
						while ( *src >= TEXT('0') && *src <= TEXT('9') )
							n = n * 10 + ((*src++) - TEXT('0'));
						fmt.precision = n;
					}
				}
				if ( *src == TEXT(',') || *src == TEXT('_') || *src == TEXT('|') )
				{
					fmt.part_ch = *src++;
					fmt.part_spc = PART_SPC_DEFAULT;
					if ( *src == TEXT('*') )
					{
						++src;
						fmt.part_spc = va_arg(arg, int);
					} else
					{
						unsigned	n;

						n = 0;
						while ( *src >= TEXT('0') && *src <= TEXT('9') )
							n = n * 10 + ((*src++) - TEXT('0'));
						if ( n > 0 )
							fmt.part_spc = n;
					}
				}
				// option
				switch ( *src )
				{
					case TEXT('l'):
						if ( src[1] == TEXT('l') )	// ll
						{
							fmt.opt = OPT_LONGLONG;
							src += 2;
						} else
						{
							fmt.opt = OPT_LONG;
							src++;
						}
						break;
					case TEXT('h'):
						fmt.opt = OPT_SHORT;
						src++;
						break;
					case TEXT('I'):
					case TEXT('i'):
						if ( src[1] == TEXT('6') && src[2] == TEXT('4') )	// i64 or I64
						{
							fmt.opt = OPT_I64;
							src += 3;
						}
						break;
				}
				// type
				fmt.type = *src;
				switch ( *src )
				{
					case TEXT('c'):
					{
						int			ch;

						src++;
						ch = va_arg(arg, int);
						*dst++ = (TCHAR) ch;
						outsize += sizeof(TCHAR);
						break;
					}
					case TEXT('d'):
					{
						TCHAR	tmp[256];
						int		sign;

						++src;
						switch ( fmt.opt )
						{
							case OPT_LONG:
							{
								long	n;

								n = va_arg(arg, long);
								if ( n < 0 )
								{
									sign = -1;
									n = -n;
								} else
									sign = 1;
								TForm::utoa((unsigned)n, tmp);
								break;
							}
							case OPT_I64:	// OPT_LONGLONG
							{
								INT64	n;

								n = va_arg(arg, INT64);
								if ( n < 0 )
								{
									sign = -1;
									n = -n;
								} else
									sign = 1;
								TForm::utoa((UINT64)n, tmp);
								break;
							}
							case OPT_SHORT:
							{
								int			n;

								n = (int) va_arg(arg, short);
								if ( n < 0 )
								{
									sign = -1;
									n = -n;
								} else
									sign = 1;
								TForm::utoa((unsigned)n, tmp);
								break;
							}
							default:
							{
								int			n;

								n = va_arg(arg, int);
								if ( n < 0 )
								{
									sign = -1;
									n = -n;
								} else
									sign = 1;
								TForm::utoa((unsigned)n, tmp);
								break;
							}
						}
						if ( sign < 0 )
							fmt.print_sign = TRUE;
						outsize += TForm::format_numstr(fmt, tmp, sign, size - outsize, dst);
						dst = buffer + outsize;
						break;
					}
					case TEXT('u'):
					{
						TCHAR	tmp[256];
						int		sign;

						++src;
						fmt.print_sign = FALSE;
						switch ( fmt.opt )
						{
							case OPT_LONG:
							{
								unsigned long	n;

								n = va_arg(arg, unsigned long);
								TForm::utoa((unsigned)n, tmp);
								break;
							}
							case OPT_I64:
							{
								UINT64	n;

								n = va_arg(arg, UINT64);
								TForm::utoa(n, tmp);
								break;
							}
							case OPT_SHORT:
							{
								unsigned	n;

								n = (unsigned) va_arg(arg, unsigned short);
								TForm::utoa(n, tmp);
								break;
							}
							default:
							{
								unsigned	n;

								n = va_arg(arg, unsigned int);
								TForm::utoa(n, tmp);
								break;
							}
						}
						sign = 0;
						outsize += TForm::format_numstr(fmt, tmp, sign, size - outsize, dst);
						dst = buffer + outsize;
						break;
					}
					case TEXT('x'):
					case TEXT('X'):
					{
						TCHAR	tmp[256];
						int		sign;
						BOOL	cap;

						++src;
						cap = fmt.type == TEXT('X') ? TRUE : FALSE;
						fmt.print_sign = FALSE;
						switch ( fmt.opt )
						{
							case OPT_LONG:
							{
								unsigned long	n;

								n = va_arg(arg, unsigned long);
								if ( n == 0 )
									fmt.num_symbol = FALSE;
								TForm::ntoahex((unsigned)n, cap, tmp);
								break;
							}
							case OPT_I64:
							{
								UINT64	n;

								n = va_arg(arg, UINT64);
								if ( n == 0 )
									fmt.num_symbol = FALSE;
								TForm::ntoahex(n, cap, tmp);
								break;
							}
							case OPT_SHORT:
							{
								unsigned	n;

								n = (unsigned) va_arg(arg, unsigned short);
								if ( n == 0 )
									fmt.num_symbol = FALSE;
								TForm::ntoahex(n, cap, tmp);
								break;
							}
							default:
							{
								unsigned	n;

								n = va_arg(arg, unsigned int);
								if ( n == 0 )
									fmt.num_symbol = FALSE;
								TForm::ntoahex(n, cap, tmp);
								break;
							}
						}
						sign = 0;
						outsize += TForm::format_numstr(fmt, tmp, sign, size - outsize, dst);
						dst = buffer + outsize;
						break;
					}
					case TEXT('k'):
					case TEXT('K'):
					{
						TCHAR	tmp[256];
						int		sign;
						INT64	n;

						++src;
						if ( fmt.precision < 0 )
							fmt.precision = 1;
						switch ( fmt.opt )
						{
							case OPT_LONG:
								n = (INT64) va_arg(arg, unsigned long);
								break;
							case OPT_I64:
								n = (INT64) va_arg(arg, UINT64);
								break;
							case OPT_SHORT:
								n = (INT64) va_arg(arg, unsigned short);
								break;
							default:
								n = (INT64) va_arg(arg, unsigned int);
								break;
						}
						if ( n < I64(0) )
						{
							sign = -1;
							n = -n;
						} else
							sign = 1;
						if ( fmt.type == TEXT('k') )
							TForm::ntostr103( n, fmt.precision, tmp);
						else
							TForm::ntostr210( n, fmt.precision, tmp);
						outsize += TForm::format_numstr(fmt, tmp, sign, size - outsize, dst);
						dst = buffer + outsize;
						break;
					}
					case TEXT('f'):
					{
						TCHAR		tmp[256], *s, *d;
						int			dec, sign, signflag;
						unsigned	len, pos;
						double		f;

						++src;
						f = va_arg(arg, double);
						if ( fmt.precision < 0 )
							fmt.precision = 6;
						s = fcvt(f, fmt.precision, &dec, &signflag);
						sign = signflag ? -1 : 0;
						d = tmp;
						len = 0;
						if ( dec <= 0 )
						{
							*d++ = TEXT('0');
							++len;
						} else if ( dec > 0 )
						{
							while ( dec-- > 0 && *s )
							{
								*d++ = *s++;
								++len;
							}
						}
						if ( fmt.precision > 0 )
						{
							*d++ = TEXT('.');
							len++;
							pos = 0;
							if ( dec < 0 )
							{
								while ( dec++ < 0 )
								{
									*d++ = TEXT('0');
									++len;
									pos++;
								}
							}
							while ( *s && (int)pos < fmt.precision )
							{
								*d++ = *s++;
								len++;
								pos++;
							}
						}
						*d = TEXT('\0');
						outsize += TForm::format_numstr(fmt, tmp, sign, size - outsize, dst);
						dst = buffer + outsize;
						break;
					}
					case TEXT('e'):
					case TEXT('E'):
					{
						TCHAR		tmp[256], *s, *d;
						int			dec, sign, signflag, e, i;
						unsigned	len, pos;
						double		f;

						++src;
						f = va_arg(arg, double);
						if ( fmt.precision < 0 )
							fmt.precision = 6;
						s = ecvt(f, fmt.precision + 1, &dec, &signflag);
						sign = signflag ? -1 : 0;
						d = tmp;
						len = pos = 0;
						while ( *s )
						{
							*d++ = *s++;
							len++;
							pos++;
							if ( pos == 1 )
							{
								*d++ = TEXT('.');
								len++;
							}
						}
						while ( (int)pos < fmt.precision + 1 )
						{
							*d++ = TEXT('0');
							len++;
							pos++;
						}
						*d++ = (TCHAR)fmt.type;	// 'E' or 'e'
						len++;
						e = dec - 1;
						if ( e < 0 )
						{
							if ( f == 0.0 )
							{
								e = 0;
								*d++ = TEXT('+');
							} else
							{
								e = -e;
								*d++ = TEXT('-');
							}
							len++;
						} else
						{
							*d++ = TEXT('+');
							len++;
						}
						for ( i = 0; i < E_FIGURE; i++ )
						{
							d[E_FIGURE-1-i] = (TCHAR)(TEXT('0') + (e % 10));
							e = e/10;
						}
						d += E_FIGURE;
						len += E_FIGURE;
						*d = TEXT('\0');
						outsize += TForm::format_numstr(fmt, tmp, sign, size - outsize, dst);
						dst = buffer + outsize;
						break;
					}
					case TEXT('s'):
					{
						const TCHAR		*text;
						unsigned		i, len, pos;

						src++;
						text = va_arg(arg, const TCHAR *);
						len = _tcslen(text);
						if ( fmt.precision < 0 )
						{
							fmt.precision = len;
						} else if ( (int)len > fmt.precision )
							len = fmt.precision;
						if ( fmt.width < (int)len )
							fmt.width = len;
						pos = 0;
						if ( fmt.left_margin == FALSE )
						{
							for ( i = 0; (int)i < fmt.width - len && outsize < size; i++ )
							{
								*dst++ = TEXT(' ');
								outsize += sizeof(TCHAR);
								pos++;
							}
						}
						for ( i = 0; (int)i < fmt.precision && i < len && outsize < size; i++ )
						{
							*dst++ = *text++;
							outsize += sizeof(TCHAR);
							pos++;
						}
						while ( (int)pos < fmt.width )
						{
							*dst++ = TEXT(' ');
							outsize += sizeof(TCHAR);
							pos++;
						}
						break;
					}
				}
			}
		} else if ( *src == '\\' )
		{
			++src;
			if ( *src == TEXT('\0') )
				break;
			switch ( *src )
			{
				case TEXT('t'):
					*src++;
					*dst++ = TEXT('\t');
					outsize += sizeof(TCHAR);
					break;
				case TEXT('r'):
					*src++;
					*dst++ = TEXT('\r');
					outsize += sizeof(TCHAR);
					break;
				case TEXT('n'):
					*src++;
					*dst++ = TEXT('\n');
					outsize += sizeof(TCHAR);
					break;
				default:
					*dst++ = *src++;
					outsize += sizeof(TCHAR);
					break;
			}
		} else
		{
#if	defined(_MBCS)
			if ( _ismbblead(*src) && _ismbbtrail(src[1]) )
			{
				if ( outsize + sizeof(TCHAR) * 2 > size )
					break;
				*dst++ = *src++;
				*dst++ = *src++;
				outsize += sizeof(TCHAR) * 2;
			} else
#endif
			{
				*dst++ = *src++;
				outsize += sizeof(TCHAR);
			}
		}
	} // end of while
	if ( outsize < size )
		*dst++ = TEXT('\0');
	return outsize;
}

int		TForm::format_numstr(TForm::Format & fmt, const TCHAR * text, int sign, unsigned size, TCHAR * buffer)
{
	TCHAR		*dst = buffer;
	const TCHAR	*s;
	TCHAR		tmp[256];
	int			i;
	unsigned	len, outlen, outsize = 0;

	if ( fmt.part_spc > 0 )
	{
		TCHAR			tmp2[256], *t;
		unsigned		top = 0, width;

		t = tmp2;
		if ( fmt.left_margin == FALSE && fmt.zero_pad )
		{
			len = _tcslen(text);
			width = len + (len-1) / fmt.part_spc;
			if ( fmt.print_sign || fmt.sign_space || sign < 0 )
				top++;
			if ( fmt.num_symbol && (fmt.type == TEXT('x') || fmt.type == TEXT('X')) )
				top += 2;
			for ( i = 0; i < fmt.width - (int)(top + width); i++ )
			{
				*t++ = TEXT('0');
				++len;
				width = len + (len-1) / fmt.part_spc;
			}
		}
		_tcscpy(t, text);
		text = TForm::seppart(fmt.part_ch, fmt.part_spc, tmp2, tmp);
	}
	len = _tcslen(text);
	outlen = 0;
	if ( fmt.print_sign || fmt.sign_space || sign < 0 )
		len++;
	if ( fmt.num_symbol && (fmt.type == TEXT('x') || fmt.type == TEXT('X')) )
		len += 2;
	if ( fmt.width < (int)len )
		fmt.width = len;
	if ( fmt.left_margin )
		fmt.zero_pad = FALSE;
	if ( fmt.zero_pad == FALSE )
	{
		if ( fmt.left_margin == FALSE )
		{
			for ( i = 0; i < fmt.width - (int)len && outsize < size; i++ )
			{
				*dst++ = TEXT(' ');
				outsize += sizeof(TCHAR);
				outlen++;
			}
		}
	}
	if ( sign < 0 )
	{
		*dst++ = TEXT('-');
		outsize += sizeof(TCHAR);
		outlen++;
	} else
	{
		if ( fmt.print_sign )
		{
			*dst++ = TEXT('+');
			outsize += sizeof(TCHAR);
			outlen++;
		} else if ( fmt.sign_space )
		{
			*dst++ = TEXT(' ');
			outsize += sizeof(TCHAR);
			outlen++;
		}
	}
	if ( fmt.num_symbol && (fmt.type == TEXT('x') || fmt.type == TEXT('X')) )
	{
		*dst++ = TEXT('0');
		outsize += sizeof(TCHAR);
		outlen++;
		*dst++ = (TCHAR)fmt.type;
		outsize += sizeof(TCHAR);
		outlen++;
	}
	if ( fmt.zero_pad )
	{
		for ( i = 0; i < (int)fmt.width - (int)len && outsize < size; i++ )
		{
			*dst++ = TEXT('0');
			outsize += sizeof(TCHAR);
			outlen++;
		}
	}
	s = text;
	while ( *s && outsize < size )
	{
		*dst++ = *s++;
		outsize += sizeof(TCHAR);
		outlen++;
	}
	while ( (int)outlen < fmt.width && outsize < size )
	{
		*dst++ = TEXT(' ');
		outsize += sizeof(TCHAR);
		outlen++;
	}
	return outsize;
}

TCHAR *	TForm::seppart(int sep, int spc, const TCHAR * text, TCHAR * buffer)
{
	unsigned	len;
	const TCHAR	*src;
	TCHAR		*dst;

	if ( (len = _tcslen(text)) == 0 )
	{
		buffer[0] = TEXT('\0');
		return buffer;
	}
	src = text + len - 1;
	dst = buffer + len + ((len-1) / spc);
	*dst-- = TEXT('\0');

	{
		unsigned	i = 0;

		for (;;)
		{
			*dst-- = *src--;
			if ( ++i >= (int)len )
				break;
			if ( (i%spc) == 0 )
				*dst-- = (TCHAR)sep;
		}
	}
	return buffer;
}

TCHAR *	TForm::utoa(UINT64 n, unsigned r, TCHAR * buf)	// static
{
	TCHAR		   *p = buf;
	int				len = 0;

	do
	{
		++len;
		*p++ = ntoa_utbl[n % (UINT64)r];
		n /= (UINT64)r;
	} while ( n > U64(0) );
	*p = TEXT('\0');

	if ( len > 1 )
	{
		TCHAR		ch;
		int			pos = 0;

		p = buf + len - 1;
		len /= 2;
		while ( pos < len )
		{
			ch       = buf[pos];
			buf[pos] = p[-pos];
			p[-pos]  = ch;
			++pos;
		}
	}
	return buf;
}

TCHAR *	TForm::utoa(UINT64 n, TCHAR * buf)	// static
{
	TCHAR		   *p = buf;
	int				len = 0;

	do
	{
		++len;
		*p++ = ntoa_utbl[n % U64(10) ];
		n /= U64(10);
	} while ( n > U64(0) );
	*p = TEXT('\0');

	if ( len > 1 )
	{
		TCHAR		ch;
		int			pos = 0;

		p = buf + len - 1;
		len /= 2;
		while ( pos < len )
		{
			ch       = buf[pos];
			buf[pos] = p[-pos];
			p[-pos]  = ch;
			++pos;
		}
	}
	return buf;
}

TCHAR *	TForm::utoa(unsigned n, TCHAR * buf)	// static
{
	TCHAR		   *p = buf;
	int				len = 0;

	do
	{
		++len;
		*p++ = ntoa_utbl[n % 10u];
		n /= 10u;
	} while ( n > 0u );
	*p = TEXT('\0');

	if ( len > 1 )
	{
		TCHAR		ch;
		int			pos = 0;

		p = buf + len - 1;
		len /= 2;
		while ( pos < len )
		{
			ch       = buf[pos];
			buf[pos] = p[-pos];
			p[-pos]  = ch;
			++pos;
		}
	}
	return buf;
}

TCHAR *	TForm::ntoahex(UINT64 n, BOOL cap, TCHAR * buf)	// static
{
	TCHAR		   *p = buf;
	int				len = 0;
	const TCHAR		*tbl = cap ? ntoa_utbl : ntoa_ltbl;

	do
	{
		++len;
		*p++ = tbl[n & 0xF];
		n >>= 4;
	} while ( n > U64(0) );
	*p = TEXT('\0');

	if ( len > 1 )
	{
		TCHAR		ch;
		int			pos = 0;

		p = buf + len - 1;
		len /= 2;
		while ( pos < len )
		{
			ch       = buf[pos];
			buf[pos] = p[-pos];
			p[-pos]  = ch;
			++pos;
		}
	}
	return buf;
}

TCHAR *	TForm::ntoahex(unsigned n, BOOL cap, TCHAR * buf)	// static
{
	TCHAR		   *p = buf;
	int				len = 0;
	const TCHAR		*tbl = cap ? ntoa_utbl : ntoa_ltbl;

	do
	{
		++len;
		*p++ = tbl[n & 0xF ];
		n >>= 4;
	} while ( n > 0u );
	*p = TEXT('\0');

	if ( len > 1 )
	{
		TCHAR		ch;
		int			pos = 0;

		p = buf + len - 1;
		len /= 2;
		while ( pos < len )
		{
			ch       = buf[pos];
			buf[pos] = p[-pos];
			p[-pos]  = ch;
			++pos;
		}
	}
	return buf;
}

TCHAR *	TForm::ntostr103(UINT64 size, int sn, TCHAR * buffer) // static
{
	unsigned			n, m;
	static UINT64		smul[] = 
	{
		U64(0)
	  , U64(10)
	  , U64(100)
	  , U64(1000)
	  , U64(10000)
	  , U64(100000)
	  , U64(1000000)
	  , U64(10000000)
	  , U64(100000000)
	  , U64(1000000000)
	  , U64(10000000000)
	  , U64(100000000000)
	};

	if ( sn >= sizeof(smul)/sizeof(*smul) )
		sn = sizeof(smul)/sizeof(*smul) - 1;
	if ( size < U64(1000) )
	{
		if ( sn > 0 )
			TForm::SPrintf(buffer, TEXT("%u.%0*uB"), (unsigned)size, sn, 0u );
		else
			TForm::SPrintf(buffer, TEXT("%uB"), (unsigned)size );
	} else if ( size < U64(1000000) )
	{	// Kilo
		n = (unsigned)(size / U64(1000));
		if ( sn > 0 )
		{
			m = (unsigned)(smul[sn] * (size % U64(1000)) / U64(1000));
			TForm::SPrintf(buffer, TEXT("%u.%0*uk"), n, sn, m);
		} else
			TForm::SPrintf(buffer, TEXT("%uk"), n );
	} else if ( size < U64(1000000000) )
	{	// Mega
		n = (unsigned)(size / U64(1000000));
		if ( sn > 0 )
		{
			m = (unsigned)(smul[sn] * (size % U64(1000000)) / U64(1000000));
			TForm::SPrintf(buffer, TEXT("%u.%0*uM"), n, sn, m);
		} else
			TForm::SPrintf(buffer, TEXT("%uM"), n);
	} else if ( size < U64(1000000000000) )
	{	// Giga
		n = (unsigned)(size / U64(1000000000));
		if ( sn > 0 )
		{
			m = (unsigned)(smul[sn] * (size % U64(1000000000)) / U64(1000000000));
			TForm::SPrintf(buffer, TEXT("%u.%0*uG"), n, sn, m);
		} else
			TForm::SPrintf(buffer, TEXT("%uG"), n );
	} else if ( size < U64(1000000000000000) )
	{	// Tela
		n = (unsigned)(size / U64(1000000000000));
		if ( sn > 0 )
		{
			m = (unsigned)(smul[sn] * (size % U64(1000000000000)) / U64(1000000000000));
			TForm::SPrintf(buffer, TEXT("%u.%0*uT"), n, sn, m);
		} else
			TForm::SPrintf(buffer, TEXT("%uT"), n);
	}

	return buffer;
}

TCHAR *	TForm::ntostr210(UINT64 size, int sn, TCHAR * buffer) // static
{
	unsigned			n, m;
	static UINT64		smul[] = 
	{
		U64(0)
	  , U64(10)
	  , U64(100)
	  , U64(1000)
	  , U64(10000)
	  , U64(100000)
	  , U64(1000000)
	  , U64(10000000)
	  , U64(100000000)
	  , U64(1000000000)
	  , U64(10000000000)
	  , U64(100000000000)
	};

	if ( sn >= sizeof(smul)/sizeof(*smul) )
		sn = sizeof(smul)/sizeof(*smul) - 1;
	if ( size < U64(0x400) )
	{
		if ( sn > 0 )
			TForm::SPrintf(buffer, TEXT("%u.%0*uB"), (unsigned)size, sn, 0u );
		else
			TForm::SPrintf(buffer, TEXT("%uB"), (unsigned)size );
	} else if ( size < U64(1000000) )
	{	// Kilo
		n = (unsigned)(size >> 10);
		if ( sn > 0 )
		{
			m = (unsigned)(smul[sn] * (size & U64(0x3FF)) >> 10);
			TForm::SPrintf(buffer, TEXT("%u.%0*uk"), n, sn, m);
		} else
			TForm::SPrintf(buffer, TEXT("%uk"), n );
	} else if ( size < U64(0x40000000) )
	{	// Mega
		n = (unsigned)(size >> 20);
		if ( sn > 0 )
		{
			m = (unsigned)(smul[sn] * (size % U64(0xFFFFF)) >> 20);
			TForm::SPrintf(buffer, TEXT("%u.%0*uM"), n, sn, m);
		} else
			TForm::SPrintf(buffer, TEXT("%uM"), n);
	} else if ( size < U64(0x10000000000) )
	{	// Giga
		n = (unsigned)(size >> 30);
		if ( sn > 0 )
		{
			m = (unsigned)(smul[sn] * (size & U64(0x3FFFFFFF)) >> 30);
			TForm::SPrintf(buffer, TEXT("%u.%0*uG"), n, sn, m);
		} else
			TForm::SPrintf(buffer, TEXT("%uG"), n );
	} else if ( size < U64(0x4000000000000) )
	{	// Tela
		n = (unsigned)(size >> 40);
		if ( sn > 0 )
		{
			m = (unsigned)(smul[sn] * (size & U64(0xFFFFFFFFFF)) >> 40);
			TForm::SPrintf(buffer, TEXT("%u.%0*uT"), n, sn, m);
		} else
			TForm::SPrintf(buffer, TEXT("%uT"), n);
	}

	return buffer;
}


#if defined(MAIN)
#include	<stdio.h>

void	Printf(const TCHAR * form, ...)
{
	va_list		arg;
	TCHAR		buffer[4096];

	va_start(arg, form);
	TForm::format(sizeof(buffer), buffer, form, arg);
	va_end(arg);
	fputs(buffer, stdout);
}

void	SPrintf(TCHAR * buffer, const TCHAR * form, ...)
{
	va_list		arg;

	va_start(arg, form);
	TForm::format(TForm::BUFFER_SIZE, buffer, form, arg);
	va_end(arg);
}

int	main(void)
{
	{
		const TCHAR	*form_tbl[] = 
		{
			"%d",

			"%10d",
			"%-10d",
			"%+-10d",

			"%3d",
			"%-3d",
			"%+-3d",

			"%010d",
			"%-010d",
			"%+-010d",

			"%03d",
			"%-03d",
			"%+-03d",
		};
		int		value_tbl[] =
		{
			0, 1, -1, 12345, -12345
		};
		int			form_no, value_no;
		TCHAR		tmp[1024], buf1[1024], buf2[1024];
		const TCHAR	*stt;

		for ( form_no = 0; form_no < sizeof(form_tbl)/sizeof(*form_tbl); form_no++ )
		{
			for ( value_no = 0; value_no < sizeof(value_tbl)/sizeof(*value_tbl); value_no++ )
			{
				sprintf(tmp, "printf(\"%s\", %d) : ", form_tbl[form_no], value_tbl[value_no]);
				sprintf(buf1, form_tbl[form_no], value_tbl[value_no]);
				SPrintf(buf2, form_tbl[form_no], value_tbl[value_no]);
				stt = strcmp(buf1, buf2) == 0 ? "OK" : "ERROR";
				printf("%-5s %s\t\"%s\"\t\"%s\"\n",  stt, tmp, buf1, buf2);
			}
		}
	}
	{
		const TCHAR	*form_tbl[] = 
		{
			"%u",

			"%10u",
			"%-10u",
			"%+-10u",

			"%3u",
			"%-3u",
			"%+-3u",

			"%010u",
			"%-010u",
			"%+-010u",

			"%03u",
			"%-03u",
			"%+-03u",
		};
		int		value_tbl[] =
		{
			0, 1, -1, 12345, -12345
		};
		int			form_no, value_no;
		TCHAR		tmp[1024], buf1[1024], buf2[1024];
		const TCHAR	*stt;

		for ( form_no = 0; form_no < sizeof(form_tbl)/sizeof(*form_tbl); form_no++ )
		{
			for ( value_no = 0; value_no < sizeof(value_tbl)/sizeof(*value_tbl); value_no++ )
			{
				sprintf(tmp, "printf(\"%s\", %u) : ", form_tbl[form_no], value_tbl[value_no]);
				sprintf(buf1, form_tbl[form_no], value_tbl[value_no]);
				SPrintf(buf2, form_tbl[form_no], value_tbl[value_no]);
				stt = strcmp(buf1, buf2) == 0 ? "OK" : "ERROR";
				printf("%-5s %s\t\"%s\"\t\"%s\"\n",  stt, tmp, buf1, buf2);
			}
		}
	}
	{
		const TCHAR	*form_tbl[] = 
		{
			"%x",

			"%10x",
			"%-10x",
			"%+-10x",

			"%3x",
			"%-3x",
			"%+-3x",

			"%010x",
			"%-010x",
			"%+-010x",

			"%03x",
			"%-03x",
			"%+-03x",

			"%#10x",
			"%#-10x",
			"%#+-10x",

			"%#3x",
			"%#-3x",
			"%#+-3x",

			"%#010x",
			"%#-010x",
			"%#+-010x",

			"%#03x",
			"%#-03x",
			"%#+-03x",
		};
		int		value_tbl[] =
		{
			0, 1, -1, 12345, -12345
		};
		int			form_no, value_no;
		TCHAR		tmp[1024], buf1[1024], buf2[1024];
		const TCHAR	*stt;

		for ( form_no = 0; form_no < sizeof(form_tbl)/sizeof(*form_tbl); form_no++ )
		{
			for ( value_no = 0; value_no < sizeof(value_tbl)/sizeof(*value_tbl); value_no++ )
			{
				sprintf(tmp,"printf(\"%s\", %u) : ", form_tbl[form_no], value_tbl[value_no]);
				sprintf(buf1, form_tbl[form_no], value_tbl[value_no]);
				SPrintf(buf2, form_tbl[form_no], value_tbl[value_no]);
				stt = strcmp(buf1, buf2) == 0 ? "OK" : "ERROR";
				printf("%-5s %s\t\"%s\"\t\"%s\"\n",  stt, tmp, buf1, buf2);
			}
		}
	}

	{
		const TCHAR	*form_tbl[] = 
		{
			"%f",

			"%10f",
			"%-10f",
			"%+-10f",

			"%3f",
			"%-3f",
			"%+-3f",

			"%010f",
			"%-010f",
			"%+-010f",

			"%03f",
			"%-03f",
			"%+-03f",
		};
		double	value_tbl[] =
		{
			0.0, 1.0, -1.0, 12345.0, -12345.0, 0.12345, -0.12345
		};
		int			form_no, value_no;
		TCHAR		tmp[1024], buf1[1024], buf2[1024];
		const TCHAR	*stt;

		for ( form_no = 0; form_no < sizeof(form_tbl)/sizeof(*form_tbl); form_no++ )
		{
			for ( value_no = 0; value_no < sizeof(value_tbl)/sizeof(*value_tbl); value_no++ )
			{
				sprintf(tmp, "printf(\"%s\", %f) : ", form_tbl[form_no], value_tbl[value_no]);
				sprintf(buf1, form_tbl[form_no], value_tbl[value_no]);
				SPrintf(buf2, form_tbl[form_no], value_tbl[value_no]);
				stt = strcmp(buf1, buf2) == 0 ? "OK" : "ERROR";
				printf("%-5s %s\t\"%s\"\t\"%s\"\n",  stt, tmp, buf1, buf2);
			}
		}
	}

	{
		const TCHAR	*form_tbl[] = 
		{
			"%e",

			"%10e",
			"%-10e",
			"%+-10e",

			"%3e",
			"%-3e",
			"%+-3e",

			"%010e",
			"%-010e",
			"%+-010e",

			"%03e",
			"%-03e",
			"%+-03e",
		};
		double	value_tbl[] =
		{
			0.0, 1.0, -1.0, 12345.0, -12345.0, 0.12345, -0.12345
		};
		int			form_no, value_no;
		TCHAR		tmp[1024], buf1[1024], buf2[1024];
		const TCHAR	*stt;

		for ( form_no = 0; form_no < sizeof(form_tbl)/sizeof(*form_tbl); form_no++ )
		{
			for ( value_no = 0; value_no < sizeof(value_tbl)/sizeof(*value_tbl); value_no++ )
			{
				sprintf(tmp, "printf(\"%s\", %f) : ", form_tbl[form_no], value_tbl[value_no]);
				sprintf(buf1, form_tbl[form_no], value_tbl[value_no]);
				SPrintf(buf2, form_tbl[form_no], value_tbl[value_no]);
				stt = strcmp(buf1, buf2) == 0 ? "OK" : "ERROR";
				printf("%-5s %s\t\"%s\"\t\"%s\"\n",  stt, tmp, buf1, buf2);
			}
		}
	}

	{
		const TCHAR	*form_tbl[] = 
		{
			"%I64d",

			"%10I64d",
			"%-10I64d",
			"%+-10I64d",

			"%3I64d",
			"%-3I64d",
			"%+-3I64d",

			"%010I64d",
			"%-010I64d",
			"%+-010I64d",

			"%03I64d",
			"%-03I64d",
			"%+-03I64d",
		};
		TForm::INT64	value_tbl[] =
		{
			I64(0), I64(1), I64(-1), I64(123456789), I64(-123456789)
		};
		int			form_no, value_no;
		TCHAR		tmp[1024], buf1[1024], buf2[1024];
		const TCHAR	*stt;

		for ( form_no = 0; form_no < sizeof(form_tbl)/sizeof(*form_tbl); form_no++ )
		{
			for ( value_no = 0; value_no < sizeof(value_tbl)/sizeof(*value_tbl); value_no++ )
			{
				sprintf(tmp, "printf(\"%s\", %d) : ", form_tbl[form_no], value_tbl[value_no]);
				sprintf(buf1, form_tbl[form_no], value_tbl[value_no]);
				SPrintf(buf2, form_tbl[form_no], value_tbl[value_no]);
				stt = strcmp(buf1, buf2) == 0 ? "OK" : "ERROR";
				printf("%-5s %s\t\"%s\"\t\"%s\"\n",  stt, tmp, buf1, buf2);
			}
		}
	}

	{
		const TCHAR	*form_tbl[] = 
		{
			"%I64u",

			"%10I64u",
			"%-10I64u",
			"%+-10I64u",

			"%3I64u",
			"%-3I64u",
			"%+-3I64u",

			"%010I64u",
			"%-010I64u",
			"%+-010I64u",

			"%03I64u",
			"%-03I64u",
			"%+-03I64u",
		};
		TForm::INT64	value_tbl[] =
		{
			I64(0), I64(1), I64(-1), I64(123456789), I64(-123456789)
		};
		int			form_no, value_no;
		TCHAR		tmp[1024], buf1[1024], buf2[1024];
		const TCHAR	*stt;

		for ( form_no = 0; form_no < sizeof(form_tbl)/sizeof(*form_tbl); form_no++ )
		{
			for ( value_no = 0; value_no < sizeof(value_tbl)/sizeof(*value_tbl); value_no++ )
			{
				sprintf(tmp, "printf(\"%s\", %d) : ", form_tbl[form_no], value_tbl[value_no]);
				sprintf(buf1, form_tbl[form_no], value_tbl[value_no]);
				SPrintf(buf2, form_tbl[form_no], value_tbl[value_no]);
				stt = strcmp(buf1, buf2) == 0 ? "OK" : "ERROR";
				printf("%-5s %s\t\"%s\"\t\"%s\"\n",  stt, tmp, buf1, buf2);
			}
		}
	}

	{
		const TCHAR	*form_tbl[] = 
		{
			"%,I64d",

			"%10,I64d",
			"%-10,I64d",
			"%+-10,I64d",

			"%3,I64d",
			"%-3,I64d",
			"%+-3,I64d",

			"%010,I64d",
			"%-010,I64d",
			"%+-010,I64d",

			"%03,I64d",
			"%-03,I64d",
			"%+-03,I64d",
		};
		TForm::INT64	value_tbl[] =
		{
			I64(0), I64(1), I64(-1), I64(123456789), I64(-123456789)
		};
		int			form_no, value_no;
		TCHAR		tmp[1024], buf2[1024];

		for ( form_no = 0; form_no < sizeof(form_tbl)/sizeof(*form_tbl); form_no++ )
		{
			for ( value_no = 0; value_no < sizeof(value_tbl)/sizeof(*value_tbl); value_no++ )
			{
				sprintf(tmp, "printf(\"%s\", %d) : ", form_tbl[form_no], value_tbl[value_no]);
				SPrintf(buf2, form_tbl[form_no], value_tbl[value_no]);
				printf("%-5s %s\t\"%s\"\t\n",  "TEST", tmp, buf2);
			}
		}
	}

	{
	//	TForm::UINT64	value_tbl[] = { U64(0), U64(1000), U64(1000000), U64(1000000000), U64(1000000000000) };
	//	TForm::UINT64	value_tbl[] = { U64(0), U64(9999), U64(9999999), U64(9999999999), U64(9999999999999) };
	//	TForm::UINT64	value_tbl[] = { U64(0), U64(0x1000), U64(0x1000000), U64(0x1000000000), U64(0x1000000000000) };
		TForm::UINT64	value_tbl[] = { U64(0), U64(0xFFFF), U64(0xFFFFFFF), U64(0xFFFFFFFFFF), U64(0xFFFFFFFFFFFFF) };
		int				value_no;
		TCHAR			buf[256];

		for ( value_no = 0; value_no < sizeof(value_tbl)/sizeof(*value_tbl); value_no++ )
		{
			TForm::ntostr103( value_tbl[value_no], 2, buf);
			Printf("%I64u = 103: %s\n", value_tbl[value_no], buf);
		}
		for ( value_no = 0; value_no < sizeof(value_tbl)/sizeof(*value_tbl); value_no++ )
		{
			TForm::ntostr210( value_tbl[value_no], 2, buf);
			Printf("%I64u = 210: %s\n", value_tbl[value_no], buf);
		}
	}

	{
		const TCHAR	*form_tbl[] = 
		{
			"%I64k",

			"%10I64k",
			"%-10I64k",
			"%+-10I64k",

			"%3I64k",
			"%-3I64k",
			"%+-3I64k",

			"%010I64k",
			"%-010I64k",
			"%+-010I64k",

			"%03I64k",
			"%-03I64k",
			"%+-03I64k",

			"%I64K",

			"%10I64K",
			"%-10I64K",
			"%+-10I64K",

			"%3I64K",
			"%-3I64K",
			"%+-3I64K",

			"%010I64K",
			"%-010I64K",
			"%+-010I64K",

			"%03I64K",
			"%-03I64K",
			"%+-03I64K",
		};
		TForm::INT64	value_tbl[] =
		{
			I64(0), I64(1), I64(-1), I64(123456789), I64(-123456789)
		};
		int			form_no, value_no;
		TCHAR		tmp[1024], buf2[1024];

		for ( form_no = 0; form_no < sizeof(form_tbl)/sizeof(*form_tbl); form_no++ )
		{
			for ( value_no = 0; value_no < sizeof(value_tbl)/sizeof(*value_tbl); value_no++ )
			{
				sprintf(tmp, "printf(\"%s\", %d) : ", form_tbl[form_no], value_tbl[value_no]);
				SPrintf(buf2, form_tbl[form_no], value_tbl[value_no]);
				printf("%-5s %s\t\"%s\"\t\n",  "TEST", tmp, buf2);
			}
		}
	}

	return 0;
}
#endif
